import argparse, sys, os
import requests, json, math
import collections, csv, time
from datetime import datetime, timedelta

from Orange import data
from orangecontrib.text.corpus import Corpus

BASE_URL = 'https://graph.facebook.com'

def progress_scale(x, subprogress):
    start = subprogress[0]
    end = subprogress[1]
    return math.ceil((start*100) + (x * (end - start)))


class FacebookCredentials:
    """ The Facebook API credentials. """
    def __init__(self, token = ''):
        self.token = token
            
    @property
    def valid(self):
        url = BASE_URL + '/facebook/?fields=id'
        headers = {'Authorization': 'Bearer ' + self.token}
        p = requests.get(url, headers=headers)    
        return 'id' in p.json().keys()

class FacebookOrangeAPI():
    attributes = []
    class_vars = []
    image_var = data.StringVariable.make("image")
    image_var.attributes["type"] = "image"
    post_metas = [(data.StringVariable('Message'), lambda doc: doc['status_message']),
             (data.DiscreteVariable('From'), lambda doc: doc['from_name']),
             (data.ContinuousVariable('likes'), lambda doc: doc['like']),
             (data.ContinuousVariable('comments'), lambda doc: doc['comments']),
             (data.ContinuousVariable('shares'), lambda doc: doc['shares']),
             (data.DiscreteVariable('top emotion'), lambda doc: doc['top_reaction']),
             (data.StringVariable('Link name'), lambda doc: doc['link_name']),
             (image_var, lambda doc: doc['picture']),
             (data.StringVariable('link'), lambda doc: doc['status_link']),
             (data.DiscreteVariable('From ID'), lambda doc: doc['from_id']),
             (data.StringVariable('Post ID'), lambda doc: doc['status_id']),
             (data.DiscreteVariable('Post type'), lambda doc: doc['status_type']),
             (data.TimeVariable('Publication Date'), lambda doc: doc['status_published']),
             (data.TimeVariable('Publication Date UTC'), lambda doc: doc['status_published_utc']),
             (data.ContinuousVariable('emotion angry'), lambda doc: doc['angry']),
             (data.ContinuousVariable('emotion love'), lambda doc: doc['love']),
             (data.ContinuousVariable('emotion haha'), lambda doc: doc['haha']),
             (data.ContinuousVariable('emotion wow'), lambda doc: doc['wow']),
             (data.ContinuousVariable('emotion sad'), lambda doc: doc['sad'])]
    text_features = [post_metas[0][0]]
    title_indices = [-1]

    def __init__(self, credentials, on_progress=None, should_break=None):
        self.utc_datecor = datetime.utcnow() - datetime.now() 
        self.pages = 0
        self.credentials = credentials
        self.on_progress = on_progress or (lambda x, y: None)
        self.should_break = should_break or (lambda: False)

    def buildUrl(self, node, version='v2.11'):
        return BASE_URL + '/' + version + '/' + node  

    def getData(self, url, params=None):
        while True:
            if self.should_break():
                return {}
            try:
                headers = {'Authorization': 'Bearer ' + self.credentials.token}
                p = requests.get(url, params=params, headers=headers)
                return p.json()
            except:                
                print('retry in 5 sec')
                for i in range(50):
                    if self.should_break():
                        return {}
                    time.sleep(0.1)

    def localToUtc(self, date):
        return date + self.utc_datecor

    def utcToLocal(self, date):
        return date - self.utc_datecor

    def processDate(self, created_time):
        return datetime.strptime(created_time,'%Y-%m-%dT%H:%M:%S+0000')

    def processStatus(self, status, engagement=True):
        d = {}
        d['status_id'] = status['id']      
        d['from_id'] = status['from']['id'] if 'from' in status.keys() else ''
        d['from_name'] = status['from']['name'] if 'from' in status.keys() else ''
        d['status_message'] = '' if 'message' not in status.keys() else status['message']
        d['status_type'] = status['type']
        d['link_name'] = '' if 'name' not in status.keys() else status['name']        

        d['status_published_utc'] = self.processDate(status['created_time'])
        d['status_published'] = self.utcToLocal(d['status_published_utc'])
        d['status_link'] = '' if 'link' not in status.keys() else status['link']
        d['picture'] = status['full_picture'] if 'full_picture' in status.keys() else ''

        topscore = 0
        d['like'] = status['like']['summary']['total_count'] if engagement else ''
        d['comments'] = status['comments']['summary']['total_count'] if engagement else ''
        d['shares'] = status['shares']['count'] if 'shares' in status.keys() else ''

        d['top_reaction'] = ''
        for score in ['love','haha','wow','sad','angry']:
            d[score] = status[score]['summary']['total_count'] if engagement else ''
            if engagement:
                d[score] = status[score]['summary']['total_count']
                if int(d[score]) > topscore:
                    topscore = int(d[score])
                    d['top_reaction'] = score
            else:
                d[score] = ''
                d['top_reaction'] = ''

        return d

    def fieldString(self, engagement=True):
        field_string = 'message,from,link,created_time,type,name,id,full_picture'
        
        if engagement:
            field_string += ',' + 'comments.limit(0).summary(true),shares.limit(0).summary(true)'
            for r in ['like','love','haha','wow','sad','angry']:
                field_string += ',' + 'reactions.type({}).limit(0).summary(true).as({})'.format(r.upper(), r.lower())
        return field_string
        

    def getStatuses(self, page_id, mode='posts', since=None, until=None, engagement=True, comments=True):
        node = page_id + '/' + mode + '/'  ## mode can be "posts" (posts by page), "feed" (all posts on page) and "tagged" (all public posts in which page is tagged
        url = self.buildUrl(node)

        params = {}
        params['fields'] = self.fieldString(engagement)
        params['limit'] = 100
        
        if since is not None: params['since'] = (self.localToUtc(since)).strftime('%Y-%m-%dT%H:%M:%S') 
        if until is not None: params['until'] = (self.localToUtc(until)).strftime('%Y-%m-%dT%H:%M:%S')
        while True:
            statuses = self.getData(url, params=params)
            if not 'data' in statuses: break

            proc_statuses = [self.processStatus(s, engagement) for s in statuses['data']]            
            yield proc_statuses            

            if not 'paging' in statuses.keys(): break
            if not 'next' in statuses['paging'].keys(): break
            url = statuses['paging']['next']

    def _search(self, page_ids, mode, since, until, max_documents, sub_progress=(0,1)):
        since = since.strftime('%Y-%m-%d')
        until = until.strftime('%Y-%m-%d')
        since = datetime.strptime(since, '%Y-%m-%d')
        until = datetime.strptime(until + 'T23:59:59', '%Y-%m-%dT%H:%M:%S')
        total_sec = float((until - since).total_seconds())
        n_pages = len(page_ids)
        
        progress_pct = 1 / float(n_pages)
        
        for page_i in range(0,n_pages):    
            page_id = page_ids[page_i].strip()
            if page_id == '': return
            if '/' in page_id: page_id = page_id.split('/')[-1]
            page_progress = progress_pct * page_i 
            n = 0
            for d in self.getStatuses(page_id, mode, since, until):
                if self.should_break():
                    return
                earliest_date = d[-1]['status_published']
                sec_to_go = (until - earliest_date).total_seconds()
                date_progress = ((sec_to_go / total_sec) * progress_pct)
                progress = math.ceil((page_progress + date_progress)*100)
                self.on_progress(progress_scale(progress, sub_progress), 100)
                for doc in d:
                    n += 1
                    if max_documents is not None:
                        if n > max_documents:
                            break
                    yield doc
                if max_documents is not None:
                    if n > max_documents:
                        break
        self.on_progress(progress_scale(100, sub_progress), 100)

    def search(self, page_ids, mode='posts', since=None, until=None, max_documents=None, sub_progress=(0,1)):
        results = []
        if since is None: since = datetime.now() - timedelta(10)
        if until is None: until = datetime.now()
        for doc in self._search(page_ids, mode, since, until, max_documents, sub_progress):
            doc['status_published'] = doc['status_published'].strftime('%Y-%m-%dT%H:%M:%S')
            doc['status_published_utc'] = doc['status_published_utc'].strftime('%Y-%m-%dT%H:%M:%S')
            results.append(doc)
             
        c = Corpus.from_documents(results, 'Facebook', self.attributes, self.class_vars, self.post_metas, self.title_indices)
        c.set_text_features(self.text_features)
        return c



    def _search_posts(self, post_ids, sub_progress=(0,1), engagement=True):
        for i, post_id in enumerate(post_ids):
            node = post_id
            url = self.buildUrl(node)

            params = {}
            params['fields'] = self.fieldString(engagement)
            params['limit'] = 100

            status = self.getData(url, params=params)
            status = self.processStatus(status)
            yield status

            progress = ((i+1) / len(post_ids)) * 100
            self.on_progress(progress_scale(progress, sub_progress), 100)
        self.on_progress(progress_scale(100, sub_progress), 100)


    def search_posts(self, post_ids, sub_progress=(0,1)):
        results = []
        for doc in self._search_posts(post_ids, sub_progress):
            doc['status_message'] = doc['status_message'].replace('\x00', '')
            doc['status_link'] = doc['status_link'].replace('\x00','')
            doc['status_published'] = doc['status_published'].strftime('%Y-%m-%dT%H:%M:%S')
            doc['status_published_utc'] = doc['status_published_utc'].strftime('%Y-%m-%dT%H:%M:%S')
            results.append(doc)

        c = Corpus.from_documents(results, 'Facebook', self.attributes, self.class_vars, self.post_metas, self.title_indices)
        c.set_text_features(self.text_features)
        return c



    def processComment(self, comment):
        has_comment_replies = 'comments' in comment.keys()
        parent = {'type': 'comment', 'comment_id': comment['id'], 'likes': comment['like']['summary']['total_count'],
                  'comment_replies': None, 'message': comment['message'], 'parent_comment_id': ''}
        parent['status_published_utc'] = self.processDate(comment['created_time'])
        parent['status_published'] = self.utcToLocal(parent['status_published_utc'])
        if has_comment_replies: parent['comment_replies'] = comment['comments']['summary']['total_count']
        yield parent

        if has_comment_replies:
            comment_replies = comment['comments']
            while True:
                for cr in comment_replies['data']:
                    child = {'type': 'comment_reply', 'comment_id': comment['id'], 'likes': cr['like']['summary']['total_count'],
                            'message': cr['message'], 'parent_comment_id': cr['id'], 'comment_replies': None}
                    child['status_published_utc'] = self.processDate(cr['created_time'])
                    child['status_published'] = self.utcToLocal(child['status_published_utc'])
                    yield child

                if not 'paging' in comment_replies.keys(): break
                if not 'next' in comment_replies['paging'].keys(): break
                url = comment_replies['paging']['next']
                comment_replies = self.getData(url)

    def _getComments(self, post_ids, comment_replies=True, sub_progress=(0,1)):
        for i, post_id in enumerate(post_ids):
            node = post_id + '/comments'
            url = self.buildUrl(node)

            params = {}
            params['fields'] = 'message,created_time,reactions.type(LIKE).summary(true).as(like)'
            if comment_replies: params['fields'] += ',comments.summary(true){message,created_time,reactions.type(LIKE).summary(true).as(like)}'
            params['limit'] = 2000

            while True:
                comments = self.getData(url, params=params)
                if len(comments['data']) == 0: break

                for comment in comments['data']:
                    for proc_comment in self.processComment(comment):
                        proc_comment['post_id'] = post_id
                        yield proc_comment

                if not 'paging' in comments.keys(): break
                if not 'next' in comments['paging'].keys(): break
                url = comments['paging']['next']
            progress = ((i+1) / len(post_ids)) * 100
            self.on_progress(progress_scale(progress, sub_progress), 100)
        self.on_progress(progress_scale(100, sub_progress), 100)


    def getComments(self, post_ids, comment_replies=True, sub_progress=(0,1)):
        attributes = []
        class_vars = []
        metas = [(data.StringVariable('Message'), lambda doc: doc['message']),
                 (data.DiscreteVariable('Type'), lambda doc: doc['type']),
                 (data.StringVariable('Post ID'), lambda doc: doc['post_id']),
                 (data.StringVariable('Comment ID'), lambda doc: doc['comment_id']),
                 (data.StringVariable('Parent comment ID'), lambda doc: doc['parent_comment_id']),
                 (data.ContinuousVariable('likes'), lambda doc: doc['likes']),
                 (data.ContinuousVariable('comment replies'), lambda doc: doc['comment_replies']),
                 (data.TimeVariable('Publication Date'), lambda doc: doc['status_published']),
                 (data.TimeVariable('Publication Date UTC'), lambda doc: doc['status_published_utc'])]
        text_features = [metas[0][0]]
        title_indices = [-1]

        results = []
        for doc in self._getComments(post_ids, comment_replies, sub_progress):
            doc['message'] = doc['message'].replace('\x00', '')
            doc['status_published'] = doc['status_published'].strftime('%Y-%m-%dT%H:%M:%S')
            doc['status_published_utc'] = doc['status_published_utc'].strftime('%Y-%m-%dT%H:%M:%S')
            results.append(doc)

        c = Corpus.from_documents(results, 'Facebook comments', attributes, class_vars, metas, title_indices)
        c.set_text_features(text_features)
        return c


#if __name__ == '__main__':
#    access_token  = ''
#    cred = FacebookCredentials(access_token)
#    f = FacebookOrangeAPI(cred)
#    for a in f.search(['volkskrant']):
#        None   



import logging
from datetime import datetime, date
from requests import HTTPError


from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QApplication, QFormLayout, QLineEdit
from Orange.data import StringVariable, TimeVariable, DiscreteVariable
from Orange.widgets import gui
from Orange.widgets.credentials import CredentialManager
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Msg
from Orange.widgets.widget import Output
from amcatclient import AmcatAPI, APIError
from orangecontrib.text.corpus import Corpus
from orangecontrib.text.widgets.utils import CheckListLayout, DatePickerInterval, ListEdit, gui_require, \
    asynchronous
from orangecontrib.text.widgets.utils.concurrent import StopExecution

DATE_OPTIONS = ["None", "Before", "After", "Between"]
DATE_NONE, DATE_BEFORE, DATE_AFTER, DATE_BETWEEN = range(len(DATE_OPTIONS))


class OWAmcat(OWWidget):
    class CredentialsDialog(OWWidget):
        name = 'The AmCAT Credentials'
        want_main_area = False
        resizing_enabled = False
        cm = CredentialManager('AmCAT orange')
        host_input = 'https://amcat.nl'
        user_input = ''
        passwd_input = ''
        token_input = ''

        class Error(OWWidget.Error):
            invalid_credentials = Msg('These credentials are invalid.')

        def __init__(self, parent):
            super().__init__()
            self.parent = parent
            self.api = None

            form = QFormLayout()
            form.setContentsMargins(5, 5, 5, 5)
            self.host_edit = gui.lineEdit(self, self, 'host_input', controlWidth=150)
            self.user_edit = gui.lineEdit(self, self, 'user_input', controlWidth=150)
            self.passwd_edit = gui.lineEdit(self, self, 'passwd_input', controlWidth=150)
            self.passwd_edit.setEchoMode(QLineEdit.Password)
            
            tokenbox = gui.vBox(self) 
            self.submit_button = gui.button(tokenbox, self, 'request new token', self.accept, width=100)
            self.token_edit = gui.lineEdit(tokenbox, self, 'token_input', controlWidth=200)
            form.addRow('Host:', self.host_edit)
            form.addRow('username:', self.user_edit)
            form.addRow('password:', self.passwd_edit) 

            form.addRow(tokenbox)  
          
            self.controlArea.layout().addLayout(form)
            self.load_credentials()

        def load_credentials(self):
            if self.cm.token:
                self.host_input, self.user_input, self.token_input = self.cm.token.split('\n')   
            
        def save_credentials(self):
            self.cm.token = '{}\n{}\n{}'.format(self.host_input, self.user_input, self.token_input)
            
        def check_credentials(self, drop_token=True):
            if drop_token:
                token = None
            else:
                token = self.token_input or None
            if token or self.passwd_input:
                try:
                    api = AmcatAPI(self.host_input, self.user_input, self.passwd_input, token)
                    if api.token is None: api = None
                except (APIError, HTTPError) as e:
                    logging.exception("Error on getting credentials")
                    api = None
            else:
                api = None
            self.passwd_input = ''
            self.token_input = api and api.token
            self.save_credentials()
            self.api = api

        def accept(self, silent=False):
            if not silent: self.Error.invalid_credentials.clear()
            
            self.check_credentials(drop_token = not silent) ## first time loading, use token from last session   
            self.parent.update_api(self.api) ## always update parent, to enable the user break the token
            if self.api:
                super().accept()
            elif not silent:
                self.Error.invalid_credentials()

    name = 'AmCAT'
    description = 'Fetch articles from The AmCAT API.'
    icon = 'icons/amcat-logo.svg'
    priority = 10

    class Outputs:
        corpus = Output("Corpus", Corpus)

    want_main_area = False
    resizing_enabled = False

    project = Setting('')
    articleset = Setting('')

    date_option = Setting(0)

    query = Setting([])
    max_documents = Setting('')


    date_from = Setting(date(1900, 1, 1))
    date_to = Setting(datetime.now().date())

    text_includes = Setting(['Headline', 'Byline', 'Content'])

    class Warning(OWWidget.Warning):
        no_text_fields = Msg('Text features are inferred when none are selected.')
        search_failed = Msg('Search failed. Try refreshing the token')

    class Error(OWWidget.Error):
        no_api = Msg('Please provide valid login information.')
        no_project = Msg('Please provide a valid (int) project id.')
        no_articleset = Msg('Please provide a valid (int) articleset id.')

    def __init__(self):
        super().__init__()
        self.corpus = None
        self.api = None
        self.output_info = ''

        # API token
        self.api_dlg = self.CredentialsDialog(self)
        self.api_dlg.accept(silent=True)
        gui.button(self.controlArea, self, 'AmCAT login',
                   callback=self.api_dlg.exec_,
                   focusPolicy=Qt.NoFocus)
        # Query
        query_box = gui.widgetBox(self.controlArea, 'Query', addSpace=True)
        aset_box = gui.hBox(query_box)
        project_edit = gui.lineEdit(aset_box, self, 'project', label='Project: ',  orientation=1, valueType=str)
        set_edit = gui.lineEdit(aset_box, self, 'articleset', label='Articleset: ', orientation=1, valueType=str)
        queryset_box = gui.hBox(query_box)
        query_box.layout().addWidget(ListEdit(self, 'query',
                         'One query per line', 80, self))

        # Year box
        def date_changed():
            d.picker_to.setVisible(self.date_option in [DATE_BEFORE, DATE_BETWEEN])
            d.picker_from.setVisible(self.date_option in [DATE_AFTER, DATE_BETWEEN])
        gui.comboBox(query_box, self, 'date_option', items=DATE_OPTIONS, label="Date filter",
                     callback = date_changed)
        date_box = gui.hBox(query_box)
        d = DatePickerInterval(date_box, self, 'date_from', 'date_to',
                               min_date=None, max_date=date.today(),
                               margin=(0, 3, 0, 0))
        date_changed()

        # Text includes features
        self.controlArea.layout().addWidget(
            CheckListLayout('Text includes', self, 'text_includes',
                            ['Headline','Byline', 'Content','Section'],
                            cols=2, callback=self.set_text_features))

        # Output
        info_box = gui.hBox(self.controlArea, 'Output')
        gui.label(info_box, self, 'Articles: %(output_info)s')

        # Buttons
        self.button_box = gui.hBox(self.controlArea)
        self.button_box.layout().addWidget(self.report_button)

        self.search_button = gui.button(self.button_box, self, 'Search',
                                        self.start_stop,
                                        focusPolicy=Qt.NoFocus)

    def update_api(self, api):
        self.Error.no_api.clear()
        self.api = api

    def new_query_input(self):
        self.search.stop()
        self.search()

    def start_stop(self):
        if self.search.running:
            self.search.stop()
        else:
            
            self.run_search()

    @gui_require('api', 'no_api')
    @gui_require('project', 'no_project')
    @gui_require('articleset', 'no_articleset')
    def run_search(self):
        if not str(self.project).isdigit(): self.project = ''
        if not str(self.articleset).isdigit(): self.articleset = ''
        if not str(self.max_documents).isdigit(): self.max_documents = ''
        self.search()
        


    @asynchronous
    def search(self):
        self.Warning.search_failed.clear()
        columns = ['id', 'date', 'medium', 'headline', 'byline', 'section', 'text','creator']
        max_documents = int(self.max_documents) if not self.max_documents == '' else None

        if not self.query and self.date_option == DATE_NONE:
            docs = self.api.get_articles(self.project, self.articleset, columns=columns, yield_pages=True)
        else:
            query = ' OR '.join(['({q})'.format(q=q) for q in self.query])
            filters = {}
            if self.date_option in [DATE_BETWEEN, DATE_AFTER]:
                filters['start_date'] = self.date_from
            if self.date_option in [DATE_BETWEEN, DATE_BEFORE]:
                filters['end_date'] = self.date_to
            docs = self.api.search(project=self.project, articleset=self.articleset, columns=columns,
                                   query=query, yield_pages=True, **filters)
        
        try:
            results = []
            for page in docs:
                results += page['results']
                try:
                    self.progress_with_info(len(results), page['total'])
                except StopExecution:
                    self.output_info = '{}/{} (interrupted)'.format(len(results), page['total'])
                    break
            if results:
                return _corpus_from_results(results)
        except APIError:
            self.Warning.search_failed()
            logging.exception("Error on searching")
        

    @search.callback(should_raise=True)
    def progress_with_info(self, n, total):
        self.output_info = '{n}/{total}'.format(**locals())
        self.progressBarSet(100 * (n / total if total else 1), None)  # prevent division by 0

    @search.on_start
    def on_start(self):
        self.Error.no_project.clear()
        self.Error.no_articleset.clear()

        self.progressBarInit(None)
        self.search_button.setText('Stop')
        self.Outputs.corpus.send(None)

    @search.on_result
    def on_result(self, result):
        self.search_button.setText('Search')
        self.progressBarFinished(None)
        self.corpus = result
        self.set_text_features()

    def set_text_features(self):
        self.Warning.no_text_fields.clear()
        if not self.text_includes:
            self.Warning.no_text_fields()

        if self.corpus is not None:
            vars_ = [var for var in self.corpus.domain.metas if var.name in self.text_includes]
            self.corpus.set_text_features(vars_ or None)
            self.Outputs.corpus.send(self.corpus)

    def send_report(self):
        self.report_items([
            ('Project', self.project),
            ('Articleset', self.articleset),
            ('Query', self.query),
            ('Date from', self.date_from),
            ('Date to', self.date_to),
            ('Text includes', ', '.join(self.text_includes)),
            ('Output', self.output_info or 'Nothing'),
        ])


CORPUS_METAS =[(StringVariable('Headline'), lambda doc: doc.get('headline', '')),
             (StringVariable('Byline'), lambda doc: doc.get('byline', '')),
             (StringVariable('Content'), lambda doc: doc.get('text', '')),
             (StringVariable('Section'), lambda doc: doc.get('section', '')),
             (StringVariable('Article_id'), lambda doc: doc['id']),
             (StringVariable('Creator'), lambda doc: doc.get('creator','')),
             (DiscreteVariable('Medium'), lambda doc: doc.get('medium', '')),
             (TimeVariable('Publication Date'), lambda doc: doc.get('date', ''))]

def _corpus_from_results(docs):
    c = Corpus.from_documents(list(docs), 'AmCAT', attributes=[], class_vars=[], metas=CORPUS_METAS, title_indices=[-1])
    return c


if __name__ == '__main__':
    app = QApplication([])
    widget = OWAmcat()
    widget.show()
    app.exec()
    widget.saveSettings()




from datetime import datetime, timedelta, date
from dateutil import relativedelta


from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QApplication, QFormLayout, QLabel, QLineEdit, QGridLayout

import Orange
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Msg
from Orange.widgets.credentials import CredentialManager
from Orange.widgets import gui
from Orange.widgets.widget import Input, Output

from orangecontrib.text.corpus import Corpus
from orangecontrib.sma.facebook_orange_api import FacebookCredentials, FacebookOrangeAPI
from orangecontrib.text.widgets.utils import CheckListLayout, QueryBox, DatePickerInterval, ListEdit,  gui_require, asynchronous

DATE_OPTIONS = ["Last week", "Last month", "Last year", "From", "Between"]
LAST_WEEK, LAST_MONTH, LAST_YEAR, DATE_FROM, DATE_BETWEEN = range(len(DATE_OPTIONS))


class OWFacebook(OWWidget):
    class CredentialsDialog(OWWidget):
        name = 'The Facebook Credentials'
        want_main_area = False
        resizing_enabled = False
        cm = CredentialManager('Facebook orange')
        app_id_input = ''
        app_secret_input = ''
        temp_token_input = ''
        token_input = ''
        
        
        class Error(OWWidget.Error):
            invalid_credentials = Msg('These credentials are invalid.')

        def __init__(self, parent):
            super().__init__()
            self.parent = parent
            self.api = None

            self.info = gui.widgetLabel(self.controlArea, 'There are two ways to connect. Either register a Facebook app or obtain a temporary access token. Both require a Facebook account.')
            self.info.setWordWrap(True);

            login_box = gui.hBox(self.controlArea) 
            login_box.setMinimumWidth(300)
            app_login = gui.widgetBox(login_box, box='Option 1: App login')
            temp_login = gui.widgetBox(login_box, box = 'Option 2: Temporary access token')
    
            ## app login
            app_form = QFormLayout()
            app_form.setContentsMargins(5, 5, 5, 5)
            app_info = gui.widgetLabel(app_login, 'To obtain an App ID and secret, <a href=\"https://developers.facebook.com/?advanced_app_create=true\">register</a> a new app (no actual app required). If you already registered an app you need to <a href=\"https://developers.facebook.com/apps/\">visit your app page</a>. The information is on the app dashboard.')
            app_info.setWordWrap(True);                
            app_info.setOpenExternalLinks(True)
            self.app_id_edit = gui.lineEdit(self, self, 'app_id_input', controlWidth=350)           
            self.app_secret_edit = gui.lineEdit(self, self, 'app_secret_input', controlWidth=350)           
            app_form.addRow('App ID:', self.app_id_edit)  
            app_form.addRow('App secret:', self.app_secret_edit)  
            app_login.layout().addLayout(app_form)
            self.submit_button = gui.button(app_login, self, 'Connect', self.app_accept)

            ## temp login
            temp_form = QFormLayout()
            temp_form.setContentsMargins(5, 5, 5, 5)
            temp_info = gui.widgetLabel(temp_login, 'To obtain a temporary (1 hour) access token, visit <a href=\"https://developers.facebook.com/tools/explorer">here</a>. Copy the token (the long line of gibberish) from the "Access Token:" box.')
            temp_info.setWordWrap(True);
            temp_info.setOpenExternalLinks(True)
            self.temp_token_edit = gui.lineEdit(self, self, 'temp_token_input', controlWidth=350)   
            temp_form.addRow('Access Token:', self.temp_token_edit)  
            temp_login.layout().addLayout(temp_form)    
            self.submit_button = gui.button(temp_login, self, 'Connect', self.temp_accept)
        
            self.load_credentials()

        def load_credentials(self):
            if self.cm.token:
                self.token_input = self.cm.token
                if '|' in self.cm.token:
                    token = self.cm.token.split('|')
                    self.app_id_input = token[0]
                    self.app_secret_input = token[1]
                else:
                    self.temp_token_input = self.cm.token
            
        def save_credentials(self):
            self.cm.token = self.token_input
            
        def check_credentials(self, drop_token=True):
            self.credentials = FacebookCredentials(self.token_input)
            self.token_input = self.credentials.token
            self.save_credentials()
            if not self.credentials.valid: self.credentials = None

        def app_accept(self):
            self.token_input = self.app_id_input + '|' + self.app_secret_input
            self.accept()

        def temp_accept(self):
            self.token_input = self.temp_token_input
            self.accept()

        def accept(self, silent=False):
            if not silent: self.Error.invalid_credentials.clear()
            
            self.check_credentials(drop_token = not silent) ## first time loading, use token from last session

            self.parent.update_api(self.credentials)
            if self.credentials:
                super().accept()
            elif not silent:
                self.Error.invalid_credentials()

    name = 'Facebook'
    description = 'Fetch articles from The Facebook Graph API.'
    icon = 'icons/facebook-logo.svg'
    priority = 10

    class Inputs:
        input_posts = Input('Facebook Posts', Orange.data.Table)

    class Outputs:
        posts = Output("Posts Corpus", Corpus)
        comments = Output("Comments Corpus", Corpus)

    want_main_area = False
    resizing_enabled = False

    page_ids = Setting([])
    modes = ['posts','feed']
    mode = Setting(0)
    comments = Setting(0)
    accumulate = Setting(0)
    max_documents = Setting('')

    input_corpus = None

    date_option = Setting(0)
    date_from = datetime.now().date()
    date_to = datetime.now().date()

    attributes = ['Message','Link name']
    text_includes = Setting(['Message'])

    class Warning(OWWidget.Warning):
        no_text_fields = Msg('Text features are inferred when none are selected.')

    class Error(OWWidget.Error):
        no_api = Msg('Please provide valid login information.')
        no_page_id = Msg('Please enter at least one page ID')
        
    def __init__(self):
        super().__init__()
        self.corpus = None
        self.api = None
        self.output_info = ''

        # API token
        self.api_dlg = self.CredentialsDialog(self)
        self.api_dlg.accept(silent=True)
        gui.button(self.controlArea, self, 'Facebook login',
                   callback=self.api_dlg.exec_,
                   focusPolicy=Qt.NoFocus)

        # Query
        self.query_box = gui.widgetBox(self.controlArea, 'Posts', addSpace=True)
        gui.label(self.query_box, self, 'Page IDs (or page URLs)')
        self.query_box.layout().addWidget(ListEdit(self, 'page_ids',
                         'One page ID per line', 80, self))

        def date_changed():
            d.picker_to.setVisible(self.date_option in [DATE_BETWEEN])
            d.picker_from.setVisible(self.date_option in [DATE_FROM, DATE_BETWEEN])
        gui.comboBox(self.query_box, self, 'date_option', items=DATE_OPTIONS, label="Date filter",
                     callback = date_changed)
        date_box = gui.hBox(self.query_box)
        d = DatePickerInterval(date_box, self, 'date_from', 'date_to',
                               min_date=None, max_date=date.today(),
                               margin=(0, 3, 0, 0))
        date_changed()

        self.mode_box = gui.widgetBox(self.query_box, box=True)
        mode_box_h = gui.hBox(self.mode_box)
        gui.radioButtonsInBox(mode_box_h, self, 'mode', btnLabels=['only posts from page itself', 'all public posts on page'], orientation=2)
        # gui.radioButtonsInBox(mode_box_h, self, 'accumulate', btnLabels=['new results', 'add to previous results'], orientation=2, label='On search:')
        gui.lineEdit(mode_box_h, self, 'max_documents', label='Max docs per page:', valueType=str, controlWidth=50)

        comments_box = gui.widgetBox(self.controlArea, box='Comments')
        gui.radioButtonsInBox(comments_box, self, 'comments', btnLabels=['ignore', 'direct comments', 'all comments'], orientation=2)


        # Text includes features
        self.query_box.layout().addWidget(
            CheckListLayout('Text includes', self, 'text_includes',
                            self.attributes,
                            cols=2, callback=self.set_text_features))

        # Output
        info_box = gui.hBox(self.controlArea, 'Output')
        gui.label(info_box, self, 'Articles: %(output_info)s')

        # Buttons
        self.button_box = gui.hBox(self.controlArea)
        self.button_box.layout().addWidget(self.report_button)

        self.search_button = gui.button(self.button_box, self, 'Search',
                                        self.start_stop,
                                        focusPolicy=Qt.NoFocus)

    def update_api(self, api):
        self.Error.no_api.clear()
        self.api = FacebookOrangeAPI(api, on_progress=self.progress_with_info,
                                       should_break=self.search.should_break)

    def new_query_input(self):
        self.search.stop()
        self.search()

    def start_stop(self):
        if self.search.running:
            self.search.stop()
        else:
            if self.api.credentials is not None and self.api.credentials.valid:
                self.Error.no_api.clear()
                self.run_search()
            else:
                self.Error.no_api()

    @gui_require('api', 'no_api')
    def run_search(self):
        if len(self.page_ids) == 0 and self.input_corpus is None:
            self.Error.no_page_id()
            return
        else:
            self.Error.no_page_id.clear()
        if not str(self.max_documents).isdigit(): self.max_documents = ''
        self.search()

    @asynchronous
    def search(self):
        mode = self.modes[self.mode]
        max_documents = int(self.max_documents) if not self.max_documents == '' else None
        
        
        #self.date_to = datetime.now().date()

        date_to = None
        date_from = None
        print(self.date_option)

        if self.date_option == LAST_WEEK:
            date_from = datetime.now().date() - timedelta(7)
        if self.date_option == LAST_MONTH:
            date_from = datetime.now().date() - relativedelta.relativedelta(months=1)
        if self.date_option == LAST_YEAR:
            date_from = datetime.now().date() - relativedelta.relativedelta(years=1)
        if self.date_option in [DATE_FROM]:
            date_from = self.date_from
        if self.date_option in [DATE_BETWEEN]:
            date_from = self.date_from
            date_to = self.date_to

        print(date_from)
        print(date_to)

        get_comments = self.comments > 0
        get_comment_replies = self.comments > 1
        posts_sub_progress = (0, 0.5) if get_comments else (0, 1)

        if self.input_corpus is None:
            posts = self.api.search(self.page_ids, mode, date_from, date_to, max_documents=max_documents,
                                    sub_progress=posts_sub_progress)
        else:
            colnames = [x.name for x in self.input_corpus.domain.metas]
            if not 'Post ID' in colnames: return (None, None)
            col_i = colnames.index('Post ID')
            post_ids = list(set(self.input_corpus.metas[:, col_i]))
            posts = self.api.search_posts(post_ids, sub_progress=posts_sub_progress)

        if get_comments:
            colnames = [x.name for x in posts.domain.metas]
            if not 'Post ID' in colnames: return (None, None)
            col_i = colnames.index('Post ID')
            post_ids = list(set(posts.metas[:, col_i]))
            comments = self.api.getComments(post_ids, get_comment_replies, sub_progress=(0.5,1))
        else:
            comments = None
        return (posts, comments)

    @search.callback(should_raise=False)
    def progress_with_info(self, n_retrieved, n_all):
        self.progressBarSet(100 * (n_retrieved / n_all if n_all else 1), None)  # prevent division by 0
        self.output_info = '{}/{}%'.format(n_retrieved, n_all)

    @search.on_start
    def on_start(self):
        self.progressBarInit()
        self.search_button.setText('Stop')
        self.Outputs.posts.send(None)
        self.Outputs.comments.send(None)

    @search.on_result
    def on_result(self, result):
        self.search_button.setText('Search')
        self.progressBarFinished(None)
        self.corpus = result[0]
        self.comments_corpus = result[1]
        self.set_text_features()


    def set_text_features(self):
        self.Warning.no_text_fields.clear()
        if not self.text_includes:
            self.Warning.no_text_fields()

        if self.corpus is not None:
            vars_ = [var for var in self.corpus.domain.metas if var.name in self.text_includes]
            self.corpus.set_text_features(vars_ or None)
            self.Outputs.posts.send(self.corpus)
        if self.comments_corpus is not None:
            self.Outputs.comments.send(self.comments_corpus)

    def send_report(self):
        self.report_items([
            ('Page IDs', self.page_ids),
            ('Date from', self.date_from),
            ('Date to', self.date_to),
            ('Text includes', ', '.join(self.text_includes)),
            ('Output', self.output_info or 'Nothing'),
        ])

    @Inputs.input_posts
    def set_data(self, input_posts):
        if input_posts is not None:
            self.input_corpus = input_posts
            self.query_box.setVisible(False)
            self.mode_box.setVisible(False)
        else:
            self.input_corpus = None
            self.query_box.setVisible(True)
            self.mode_box.setVisible(True)


if __name__ == '__main__':
    app = QApplication([])
    widget = OWFacebook()
    widget.show()
    app.exec()
    widget.saveSettings()




import re

from AnyQt.QtWidgets import QApplication, QGridLayout, QLabel, QLineEdit, QSizePolicy, QScrollArea, QCheckBox
from AnyQt.QtCore import Qt, QTimer, QSize
import Orange
from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output, Msg

class OWDictionary(OWWidget):
    name = "Dictionary"
    description = "Create a dictionary"
    icon = "icons/dictionary.svg"
    priority = 10

    want_main_area = False
    resizing_enabled = True
    
    queries = Setting([["", ""], ["", ""]])
    querytable = None
    querytable_attr = []
    querytable_metas = []
    querytable_vars = []
    querytable_vars2 = []
    label_in = [None]
    query_in = [None]
    weight_in = [None]
    add_quotes = Setting(True)
    sync = Setting(False)
    send = Setting(False)


    class Inputs:
        data = Input("Table", Orange.data.Table)

    class Outputs:
        dictionary = Output("Dictionary", Orange.data.Table)

    class Error(OWWidget.Error):
        no_query = Msg('You need to select a query column to import, select or sync a dictionary.')

    def __init__(self):
        super().__init__()

        self.query_edits = []
        self.remove_buttons = []
        
        # GUI
        scrollArea = QScrollArea()
       

        #### header
        head_box = gui.hBox(self.controlArea)
        head_box.setMaximumHeight(200)
        
        info_box = gui.widgetBox(head_box, 'Info')
        self.info = gui.widgetLabel(info_box,
           ("Import and/or create queries\n\n" +
            "If a Table with a dictionary is connected as input,\n" +
            "it can be imported, in which case multiple query\n" +
            "terms with the same label will be combined.\n\n" +
            "If the label is a number, or if negative weights are\n" +
            "used (e.g. sentiment dictionaries), the query will be\n" +
            "split into positive and negative parts.\n\n"
            "If a row in the dictionary contains multiple terms\n" +
            "separated with spaces, it will be seen as a phrase\n" +
            "(by adding quotes). This can be turned of in case\n" +
            "rows are already boolean queries."))

        ## from input
        input_box = gui.widgetBox(head_box, "Create queries from dictionary file")
        input_box.setMaximumWidth(350)

        gui.button(input_box, self, 'multiple words are phrases', toggleButton=True, value='add_quotes',
                   buttonType=QCheckBox)
        inputline_box = gui.hBox(input_box)
        inputline_box.setMinimumHeight(70)
        gui.listBox(inputline_box, self, 'query_in', labels='querytable_vars', box = 'Query column', callback=self.update_if_sync)
        gui.listBox(inputline_box, self, 'label_in', labels='querytable_vars2', box = 'Label column', callback=self.update_if_sync)
        self.weight_select = gui.listBox(inputline_box, self, 'weight_in', labels='querytable_vars2', box = 'Weight column', callback=self.update_if_sync)

        input_button_box = gui.hBox(input_box)
        gui.button(input_button_box, self, 'Keep synchronized', self.sync_on_off, toggleButton=True, value='sync', buttonType=QCheckBox)
        gui.button(input_button_box, self, 'Import', self.import_queries)
        gui.button(input_button_box, self, 'Append', self.append_queries)

        ## query field
        query_box = gui.widgetBox(self.controlArea)
        
        self.queries_box = QGridLayout()
        query_box.layout().addLayout(self.queries_box)

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scroll.setWidget(query_box)

        querygridbox = gui.widgetBox(self.controlArea, 'Query')
        querygridbox.layout().addWidget(scroll)       
        
        querygridbox.setMinimumHeight(200)
        querygridbox.setMinimumWidth(500)
        gui.rubber(query_box)

        self.queries_box.setColumnMinimumWidth(0, 5)
        self.queries_box.setColumnMinimumWidth(1, 60)
        self.queries_box.setColumnMinimumWidth(2, 350)
        self.queries_box.setColumnStretch(1, 0)
        self.queries_box.setColumnStretch(2, 100)
        self.queries_box.addWidget(QLabel("Label"), 0, 1)
        self.queries_box.addWidget(QLabel("Query"), 0, 2)
        self.update_queries()

        gui.button(query_box, self, "add query", callback=self.add_row, autoDefault=False)

        ## buttons
        scarybuttonbox = gui.hBox(self.controlArea)
        scarybuttonbox.layout().setAlignment(Qt.AlignRight)
        gui.button(scarybuttonbox, self, "remove all queries", callback=self.remove_all, width=150)

        QTimer.singleShot(0, self.send_queries) ## for send on startup

    def update_if_sync(self):
        if self.sync:
            self.import_queries()
    
    def query_changed(self):
        self.sync = False
        self.send_queries()
            
    def get_queries(self):
        for l, q in self.query_edits:
            l = l.text()
            l = re.sub('[#?]', '', l)
            q = q.text()
            yield [l, q]

    def send_queries(self):
        #if self.send:
        self.queries = list(self.get_queries())
        self.update_queries()
        valid_queries = [[label, query] for label, query in self.queries if not query == '']
        self.send_dictionary(valid_queries) 

    def send_dictionary(self, queries):
        domain = Orange.data.Domain([], metas = [
                                     Orange.data.StringVariable("label"),
                                     Orange.data.StringVariable("query")])
        out = Dictionary(domain, queries)
        self.Outputs.dictionary.send(out)

    def send_output(self, queries):
        domain = Orange.data.Domain([], metas = [
                                 Orange.data.StringVariable("label"),
                                 Orange.data.StringVariable("query")])

    def adjust_n_query_rows(self):
        def _add_line():
            self.query_edits.append([])
            n_lines = len(self.query_edits)
                            
            label_edit = gui.LineEditWFocusOut(self, self.query_changed)
            label_edit.setMaxLength(500000)
            self.query_edits[-1].append(label_edit)
            self.queries_box.addWidget(label_edit, n_lines, 1)

            query_edit = gui.LineEditWFocusOut(self, self.query_changed)
            query_edit.setMaxLength(500000)
            self.query_edits[-1].append(query_edit)
            self.queries_box.addWidget(query_edit, n_lines, 2)

            button = gui.button(
                None, self, label='×', flat=True, height=20,
                styleSheet='* {font-size: 16pt; color: silver}'
                           '*:hover {color: black}',
                autoDefault=False, callback=self.remove_row)
            button.setMinimumSize(QSize(3, 3))
            self.remove_buttons.append(button)
            self.queries_box.addWidget(button, n_lines, 0)
            for coli, kwargs in enumerate(
                    (dict(alignment=Qt.AlignRight),
                     dict(alignment=Qt.AlignLeft, styleSheet="color: gray"))):
                label = QLabel(**kwargs)
                self.queries_box.addWidget(label, n_lines, 3 + coli)

        def _remove_line():
            for edit in self.query_edits.pop():
                edit.deleteLater()
            self.remove_buttons.pop().deleteLater()

        n = len(self.queries)
        while n > len(self.query_edits):
            _add_line()
        while len(self.query_edits) > n:
            _remove_line()

    def add_row(self):
        self.queries.append(["", ""])
        self.adjust_n_query_rows()

    def remove_row(self):
        remove_idx = self.remove_buttons.index(self.sender())
        del self.queries[remove_idx]
        self.update_queries()

    def remove_all(self):
        self.queries = []
        self.update_queries()
       
    def update_queries(self):
        self.adjust_n_query_rows()
        self.queries_to_edits()

    def queries_to_edits(self):
        for editr, textr in zip(self.query_edits, self.queries):
            for edit, text in zip(editr, textr):
                edit.setText(text)

    def sync_on_off(self):
        valid_input = self.query_in is not None
        if self.sync and valid_input:
            self.Warning.no_query.clear()
            self.import_queries()
            self.sync = True ## ugly workaround around for unsetting sync with append_queries
            #self.send = True
            self.send_queries()
        else:
            self.Error.no_query()
            #self.send=False
            self.sync = False ## disable synchronize if label_in or query_in are not specified


    def import_queries(self):
        self.queries = []
        self.append_queries()
        
    def append_queries(self):
        self.sync = False
        self.send = False
        label_col = self.querytable_vars2[self.label_in[0]] if not self.label_in[0] is None else None
        query_col = self.querytable_vars[self.query_in[0]] if not self.query_in[0] is None else None
        weight_col = self.querytable_vars2[self.weight_in[0]] if not self.weight_in[0] is None else None
        if label_col == '[not used]': label_col = None
        if weight_col == '[not used]': weight_col = None

        if self.querytable is not None and query_col is not None:
            self.Error.no_query.clear()
            add_queries = self.querytable.import_dictionary(label_col, query_col, weight_col, self.add_quotes)
            self.queries = self.queries + add_queries
            self.update_queries()
        else:
            self.Error.no_query()
        self.send_queries()

    @Inputs.data
    def set_data(self, data):
        if data is not None:
            self.querytable = Dictionary(data)
            self.querytable_attr = self.querytable.attrnames()
            self.querytable_meta = self.querytable.metanames()
            self.querytable_vars = self.querytable_attr + self.querytable_meta
            self.querytable_vars2 = ['[not used]'] + self.querytable_vars
        else:
            self.querytable = None
            self.querytable_attr, self.querytable_metas, self.querytable_vars, self.querytable_vars2 = [], [], [], []
            self.label_in, self.query_in = [None], [None]



class Dictionary(Orange.data.Table):
    """Internal class for storing a dictionary."""

    def attrnames(self):
        return [x.name for x in self.domain.attributes]

    def metanames(self):
        return [x.name for x in self.domain.metas]

    def get_column(self, name):
        for i, a in enumerate(self.attrnames()):
            if a == name:   
                return [str(v[0]) for v in self[:,i]]   
        for i, m in enumerate(self.metanames()):
            if m == name:
                return self.metas[:,i]  ## metas is in numpy format, so this already returns simple list

    def get_dictionary(self, label_col='label', query_col='query'):
        label = self.get_column(label_col)
        query = self.get_column(query_col)
        return [list(a) for a in zip(label,query)]

    def import_dictionary(self, label_col='label', query_col='query', weight_col='weight', add_quotes=True):
        query = self.get_column(query_col)
        #if label_col is None: label_col = weight_col  ## special case where weight is sentiment score
        label = self.get_column(label_col) if label_col is not None else ["no label"] * len(query)
        weight = self.get_column(weight_col) if weight_col is not None else [1] * len(query)

        qdict = {}
        for l, q, w in zip(label, query, weight):
            q = clean_query(q)
            if can_float(l):
                l = 'positive' if float(l) > 0 else 'negative'
            if add_quotes:
                if not '"' in q and len(q.split()) > 1:
                    q = '"' + q + '"'
            else:
                if len(q.split()) > 1:
                    q = '(' + q + ')'
            if can_float(w):
                w = float(w)
                if not w == 1:
                    if w < 0 and not l == 'negative':
                        l = l + ' (negative)'
                    ## regex to add weight to (phrases wrapped in quotes) or (single terms except OR/AND/NOT)
                    q = re.sub(r'(\"[^\"]*\")|([^ \(\)(OR)(AND)(NOT)]+)', r'\1\2^' + str(abs(w)), q)

            qdict[l] = q if not l in qdict.keys() else qdict[l] + ' OR ' + q

        return [[k, v] for k,v in qdict.items()]


def can_float(x):
    try:
        float(x)
        return(True)
    except ValueError:
        return(False)

def clean_query(q):
    q = re.sub('”|“', '"', q)
    return q

if __name__ == '__main__':
    app = QApplication([])
    widget = OWDictionary()
    widget.show()
    app.exec()
    widget.saveSettings()



import multiprocessing
from tempfile import TemporaryDirectory
from threading import Lock
from collections import defaultdict

from progressmonitor import monitored, ProgressMonitor
from whoosh import scoring
from whoosh.analysis.tokenizers import SpaceSeparatedTokenizer
from whoosh.index import create_in
from whoosh.fields import *
from whoosh.qparser.default import QueryParser
from orangecontrib.text.corpus import Corpus

_GLOBAL_LOCK = Lock()

class Index(object):
    """Wrapper around a whoosh index"""

    def __init__(self, corpus, procs=2, limitmb=256):
        self.tokens = corpus.tokens
        self.tempdir = TemporaryDirectory(prefix="orange3sma_index")
        schema = Schema(text=TEXT(stored=False, analyzer=SpaceSeparatedTokenizer(), phrase=True), doc_i=NUMERIC(int, 64, signed=False, stored=True))
        self.index = create_in(self.tempdir.name, schema)
        w = self.index.writer(limitmb=limitmb, procs=procs, multisegment=(procs > 1))
        for doc_i, doc_tokens in enumerate(self.tokens):
            w.add_document(text=doc_tokens, doc_i=doc_i)
        w.commit()
        
    def search(self, query: str, frequencies=False):
        """
        Get the indices of the documents matching the query
        :param query: The whoosh query string
        :param frequencies: If true, return pairs of (docnum, frequency) rather than only docnum
        :return: sequence of document numbers (and freqs, if frequencies is True)
        """

        with self.index.searcher(weighting=scoring.Frequency) as searcher:
            if frequencies:
                ## for some reason, using searcher.search counts all individual occurrences of the terms in a phrase ("term1 term2")
                ## after the phrase occurs at least once. So for frequencies, we use this lengthy alternative
                ## (I expect that somewhere a setting is hidden to simply fix this with searcher.search, but no clue yet)
                results = defaultdict(lambda:float(0))
                queries = divide_query(query)

                for i, q in enumerate(queries):
                    q = QueryParser("text", self.index.schema).parse(q)
                    matcher = q.matcher(searcher)

                    while matcher.is_active():
                        docnum = searcher.reader().stored_fields(matcher.id())['doc_i']
                        bd = boostdict(matcher)
                        for s in matcher.spans():
                            results[docnum] += bd[s] if s in bd else 1
                        matcher.next()
                return [(k,v) for k,v in results.items()]
            else:
                query = QueryParser("text", self.index.schema).parse(query)
                results = searcher.search(query, limit=None, scored=False, sortedby=None)
                return [results[i]['doc_i'] for i in range(len(results))]

    def get_context(self, query: str, window: int = 30):
        """
        Get the words in the context (n-word window) of all locations of the string

        :param query: search query
        :param window: window size (in words)
        :return: a generator of (id, text) pairs
        """
        def get_window_tokens(tokens, spans):
            position = -1
            for span in spans:
                for position in range(max(position + 1, span.start - window), min(len(tokens), span.end + window + 1)):
                    yield tokens[position]

        query = QueryParser("text", self.index.schema).parse(query)
        with self.index.searcher() as searcher:
            matcher = query.matcher(searcher)
            while matcher.is_active():
                docnum = searcher.reader().stored_fields(matcher.id())['doc_i']
                yield docnum, list(get_window_tokens(self.tokens[docnum], matcher.spans()))
                matcher.next()

    def term_statistics(self):
        """
        Yield the term and document frequency for each term in the index
        :return: generator of (term, termfreq, docfreq) triples
        """
        with self.index.reader() as r:
            for t in r.field_terms('text'):
                yield t, r.doc_frequency('text', t), r.frequency('text', t)

    def reader(self, *args, **kargs):
        return self.index.reader(*args, **kargs)


@monitored(100, "Indexing corpus")
def get_index(corpus: Corpus, monitor: ProgressMonitor, multiple_processors=False, **kargs) -> Index:
    """
    Get the index for the provided corpus, reindexing (and tokenizing) if needed
    """
    with _GLOBAL_LOCK:
        if not hasattr(corpus, "_orange3sma_index_lock"):
            corpus._orange3sma_index_lock = Lock()
    with corpus._orange3sma_index_lock:
        ix = getattr(corpus, "_orange3sma_index", None)
        if not (ix and ix.tokens is corpus._tokens):
            monitor.update(0, "Getting tokens")
            corpus.tokens  # force tokens
            procs = max(1, multiprocessing.cpu_count()-1) if multiple_processors else 1
            monitor.update(50, "Creating index")
            ix = Index(corpus, procs=procs, **kargs)
            corpus._orange3sma_index = ix
    return ix


def divide_query(query):
    """
    divide query into parts connected by OR statements (that can be executed separately)
    necessary because for long queries, whoosh uses an ArrayUnionMatcher which messes up
    the frequency calculations
    """
    queries = ['']
    querylist = re.findall('\([^\)]*\)|\S+', query)
    for q in querylist:
        if q == 'OR':
            queries.append('')
            continue
        else:
            if queries[-1]: queries[-1] += ' '
            queries[-1] += q
    return queries

def boostdict(m, bd=None):
    """
    Get a dictionary with boost (i.e. weight) scores (dict values) for different spans (dict keys)
    """
    if bd is None:
        bd = {}
    if m.is_active():
        if hasattr(m, 'boost'):
            for s in m.spans():
                bd[s] = float(m.boost)
                return bd
        if hasattr(m, 'children'):
            children = m.children()
            for child in m.children():
                bd = boostdict(child, bd)
    return bd

if __name__ == '__main__':
    from Orange import data
    from orangecontrib.text.corpus import Corpus

    metas = [(data.StringVariable('headline'), lambda doc: doc['headline']),
             (data.StringVariable('text'), lambda doc: doc['text']),
             (data.StringVariable('id'), lambda doc: doc['id'])]
    text_features = [metas[0][0], metas[1][0]]
    title_indices = [-1]
    d = [{'headline': 'titel_a', 'text': 'x x y this tests a test y x x', 'id': '1'},
         {'headline': 'titel_b', 'text': 'more tests!!!', 'id': '2'}]
    c = Corpus.from_documents(d, 'example', [], [], metas, title_indices)
    c.set_text_features(text_features)

    _ix = Index(c)

    import numpy as np
    terms = list(_ix.term_statistics())
    words = np.array([["a"], ["b"]])
    print(words)




from collections import OrderedDict
from typing import Mapping, Tuple

import numpy as np
from Orange.data.domain import Domain
from Orange.data.table import Table
from Orange.data.variable import StringVariable, ContinuousVariable
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.widget import OWWidget
from Orange.widgets import gui
from orangecontrib.text.widgets.utils.concurrent import asynchronous
from progressmonitor import monitored, ProgressMonitor

from orangecontrib.text import Corpus


def _create_table(words, scores: Mapping[str, np.array]) -> Table:
    """
    Create an Orange table from the word scores
    :param words: list of words
    :param scores: mapping of {label: score_array}. Use ordereddict to preserve column order
    :return: a Table object
    """
    values = list(scores.values())
    order = (-values[0]).argsort()
    data = np.column_stack(values)[order]
    words = np.array(words).reshape(len(words), 1)[order]
    domain = Domain([ContinuousVariable(label) for label in scores],
                    metas=[StringVariable("term")])
    return Table(domain, data, metas=words)


@monitored(100)
def get_counts(corpus: Corpus, monitor: ProgressMonitor) -> Tuple[Mapping[str, int], Mapping[str, int]]:
    monitor.update(0, "Getting tokens")
    tokens = corpus.tokens  # forces tokens to be created
    n = len(tokens)
    tf, df = {}, {}  # tf: {word : freq}, df: {word: {doc_i, ...}}
    monitor.update(50, "Counting words")
    with monitor.subtask(50) as sm:
        sm.begin(n)
        for i, doc_tokens in enumerate(tokens):
            sm.update(message="Counting words {i}/{n}".format(**locals()))
            for t in doc_tokens:
                df.setdefault(t, set()).add(i)
                tf[t] = tf.get(t, 0) + 1
        df = {w: len(df[w]) for w in df}
        return tf, df


def _relfreq(c):
    c2 = c+1
    return c2/c2.sum()


@monitored(100)
def compare(corpus: Corpus, reference_corpus: Corpus, monitor: ProgressMonitor):
    tf1, df1 = get_counts(corpus, monitor.submonitor(40))
    tf2, df2 = get_counts(reference_corpus, monitor.submonitor(40))

    words = list(set(df1.keys()) | set(df2.keys()))
    counts = np.fromiter((tf1.get(t, 0) for t in words), int)
    docfreqs = np.fromiter((df1.get(t, 0) for t in words), int)
    refcounts = np.fromiter((tf2.get(t, 0) for t in words), int)
    refdocfreqs = np.fromiter((df2.get(t, 0) for t in words), int)

    relc, relcr = _relfreq(counts), _relfreq(refcounts)
    over = relc / relcr
    return _create_table(words, OrderedDict([
            ("percent", relc),
            ("frequency", counts),
            ("docfreq", docfreqs),
            ("overrepresentation", over),
            ("reference_percent", relcr),
            ("reference_frequency", refcounts),
            ("reference_docfreq", refdocfreqs),
       ]))


@monitored(100)
def frequencies(corpus, monitor):
    tf, df = get_counts(corpus, monitor.submonitor(90))

    words = list(tf.keys())
    counts = np.fromiter((tf.get(t, 0) for t in words), int)
    docfreqs = np.fromiter((df.get(t, 0) for t in words), int)
    reldocfreqs = _relfreq(counts)

    monitor.update(10)
    return _create_table(words, OrderedDict([
        ("frequency", counts),
        ("docfreq", docfreqs),
        ("relative_docfreq", reldocfreqs),
    ]))


class OWCorpusStatistics(OWWidget):
    name = "Corpus Statistics"
    description = "Calculate word frequencies and optionally compare to reference corpus"
    icon = "icons/stats.svg"
    priority = 10

    want_main_area = False
    resizing_enabled = False

    class Inputs:
        data = Input("Corpus", Corpus)
        reference = Input("Reference Corpus", Corpus)

    class Outputs:
        statistics = Output("Statistics", Table)

    class Error(OWWidget.Error):
        pass

    def __init__(self):
        super().__init__()

        self.corpus = None
        self.reference_corpus = None

        box = gui.widgetBox(self.controlArea, "Info")
        self.info = gui.widgetLabel(box, 'Output to Data Table widget to view results')

    @asynchronous
    def calculate(self):
        with ProgressMonitor().task(100, 'Calculating statistics..') as monitor:
            monitor.add_listener(self.callback)
            if self.reference_corpus:
                t = compare(self.corpus, self.reference_corpus, monitor=monitor)
            else:
                t = frequencies(self.corpus, monitor=monitor)
            return t


    @calculate.on_result
    def on_result(self, result):
        self.progressBarFinished()
        self.Outputs.statistics.send(result)

    @calculate.callback(should_raise=True)
    def callback(self, monitor):
        self.progressBarSet(monitor.progress * 100)

    @calculate.callback(should_raise=True)
    def progress_with_info(self, n, total):
        self.output_info = '{n}/{total}'.format(**locals())
        self.progressBarSet(100 * (n / total if total else 1), None)  # prevent division by 0

    @calculate.on_start
    def on_start(self):
        self.progressBarInit()

    def go(self):
        if self.corpus is None:
            self.Outputs.statistics.send(None)
        else:
            self.calculate()

    @Inputs.data
    def set_data(self, corpus):
        self.corpus = corpus
        self.go()

    @Inputs.reference
    def set_refdata(self, reference_corpus):
        self.reference_corpus = reference_corpus
        self.go()


if __name__ == '__main__':
    from Orange import data
    from orangecontrib.text.corpus import Corpus

    def corpus(*texts):
        metas = [(data.StringVariable('text'), lambda doc: doc['text'])]
        d = [{'text': t} for t in texts]
        c = Corpus.from_documents(d, 'example', [], [], metas, [-1])
        c.set_text_features([metas[0][0]])
        return c
    c1 = corpus("dit is een test", "met een document")
    c2 = corpus("test", "en nog een document")
    print(compare(c1, c2))



import numpy as np
import re

import progressmonitor
from AnyQt.QtCore import Qt
from AnyQt.QtGui import QIntValidator, QColor
from AnyQt.QtWidgets import QApplication, QCheckBox

from Orange.widgets import gui
from Orange.widgets.settings import Setting
from Orange.widgets.widget import OWWidget, Input, Output, Msg
from orangecontrib.text.corpus import Corpus
from orangecontrib.text.widgets.utils.concurrent import asynchronous
from orangecontrib.text.widgets.utils.widgets import ListEdit
from progressmonitor import ProgressMonitor

from orangecontrib.sma.index import get_index
from orangecontrib.sma.widgets.OWDictionary import Dictionary


def parse_query(string):
    m = re.match("([^#]*\w)#(.*)", string)
    l, q = m.groups() if m else [string, string]
    return l.strip(), q.strip()


QUERY_MODES = ['count', 'filter']

class OWQuerySearch(OWWidget):
    name = "Query Search"
    description = "Subset a Corpus based on a query"
    icon = "icons/queryfilter.svg"
    priority = 10

    want_main_area = False
    resizing_enabled = False

    queries = Setting('')
    dictionary_text = Setting('')
    include_unmatched = Setting(False)
    context_window = Setting('')
    dictionary_on = False
    window_disabled_text = ''

    query_mode = Setting(0)

    class Inputs:
        data = Input("Corpus", Corpus)
        dictionary = Input("Dictionary", Dictionary)

    class Outputs:
        sample = Output("Filtered Corpus", Corpus)
        remaining = Output("Unselected Documents", Corpus)

    class Error(OWWidget.Error):
        no_query = Msg('Please provide a query.')

    def __init__(self):
        super().__init__()

        self.corpus = None

        # GUI
        box = gui.widgetBox(self.controlArea, "Info")
        self.info = gui.widgetLabel(box, 'Connect an input corpus to start querying')

        self.import_box = gui.vBox(self.controlArea, 'Dictionary')
        self.import_box.setVisible(False)
        gui.button(self.import_box, self, 'Use dictionary', toggleButton=True, value='dictionary_on', buttonType=QCheckBox)
        self.dictionarybox = ListEdit(self, 'dictionary_text', '', 60, self)
        self.dictionarybox.setTextColor(QColor(100, 100, 100))
        self.dictionarybox.setReadOnly(True)
        self.import_box.layout().addWidget(self.dictionarybox)

        query_box = gui.widgetBox(self.controlArea, 'Query', addSpace=True)
        self.querytextbox = ListEdit(self, 'queries', '', 80, self)
        query_box.layout().addWidget(self.querytextbox)

        query_parameter_box = gui.hBox(self.controlArea, self)
        gui.radioButtonsInBox(query_parameter_box, self, 'query_mode', btnLabels=QUERY_MODES, box="Query mode", callback=self.toggle_mode)

        self.count_mode_parameters = gui.vBox(query_parameter_box, self)
        gui.checkBox(self.count_mode_parameters, self, 'include_unmatched', label="Include unmatched documents")

        self.filter_mode_parameters = gui.widgetBox(query_parameter_box, self)
        gui.lineEdit(self.filter_mode_parameters, self, "context_window", 'Output words in context window',
                     validator=QIntValidator())

        self.toggle_mode()

        info_box = gui.hBox(self.controlArea, 'Status')
        self.status = 'Waiting for input'
        gui.label(info_box, self, '%(status)s')

        self.search_button = gui.button(self.controlArea, self, 'Search',
                                        self.start_stop,
                                        focusPolicy=Qt.NoFocus)

    def run_search(self):
        if not self.dictionary_on and not self.queries:
            self.Error.no_query()
        else:
            self.Error.no_query.clear()
        if self.corpus is None:
            self.info.setText('Connect an input corpus to start querying')
            self.Outputs.sample.send(None)
        else:
            # start async search
            self.search()

    def start_stop(self):
        if self.search.running:
            self.search.stop()
        else:
            self.run_search()

    def import_dictionary(self):
        self.dictionary_text = []
        for label, query in self.dictionary:
            q = label.strip() + '# ' + query.strip() if label else query.strip()
            self.dictionary_text.append(q)
        self.dictionarybox.setText('\n'.join(self.dictionary_text))
        self.Error.no_query.clear()

    def toggle_mode(self):
        if QUERY_MODES[self.query_mode] == 'count':
            self.filter_mode_parameters.setVisible(False)
            self.count_mode_parameters.setVisible(True)
        else:
            self.filter_mode_parameters.setVisible(True)
            self.count_mode_parameters.setVisible(False)

    @asynchronous
    def search(self):
        indices = [0]
        queries = self.queries
        if self.dictionary_on and type(self.dictionary_text) is list:
            if type(queries) is list:  ## queries starts as str, but becomes list if queries are given (don't ask)
                queries = queries + self.dictionary_text
            else:
                queries = self.dictionary_text

        with ProgressMonitor().task(100, 'Starting search..') as monitor:
            monitor.add_listener(self.callback)
            index = get_index(self.corpus, monitor=monitor.submonitor(50))

            if QUERY_MODES[self.query_mode] == 'filter':
                # simple search
                query = " OR ".join('({})'.format(parse_query(q)[1]) for q in queries)
                if not self.context_window:
                    selected = list(index.search(query))
                    sample = self.corpus[selected]
                else:
                    sample = self.corpus.copy()
                    sample._tokens = sample._tokens.copy()
                    selected = []
                    for i, context in index.get_context(query, int(self.context_window)):
                        sample._tokens[i] = context
                        selected.append(i)
                    sample = sample[selected]
                o = np.ones(len(self.corpus))
                o[selected] = 0
                remaining = np.nonzero(o)[0]
                remaining = self.corpus[remaining]
            else:
                sample = self.corpus.copy()
                remaining = None
                seen = set()
                for q in queries:
                    label, q = parse_query(q)
                    # todo: implement as sparse matrix!
                    scores = np.zeros(len(sample), dtype=np.float)

                    for i, j in index.search(q, frequencies=True):
                        seen.add(i)
                        scores[i] = j
                    scores = scores.reshape((len(sample), 1))
                    sample.extend_attributes(scores, [label])
                if self.include_unmatched:
                    remaining = None
                else:
                    selected = list(seen)
                    o = np.ones(len(self.corpus))
                    o[selected] = 0
                    remaining = np.nonzero(o)[0]
                    remaining = self.corpus[remaining]
                    sample = sample[selected]
            return sample, remaining

    @search.callback(should_raise=True)
    def callback(self, monitor):
        self.progressBarSet(monitor.progress * 100)
        self.status = monitor.message

    @search.on_start
    def on_start(self):
        self.progressBarInit()

    @search.on_result
    def on_result(self, result):
        self.progressBarFinished()
        if result:
            sample, remaining = result
            self.info.setText('%d sampled instances' % len(sample))
        else:
            sample, remaning = None, None
            self.info.setText('(no input)')
        self.Outputs.sample.send(sample)
        self.Outputs.remaining.send(remaining)

    @Inputs.data
    def set_data(self, corpus):
        self.corpus = corpus
        self.run_search()

    @Inputs.dictionary
    def set_dictionary(self, dictionary):
        if dictionary:
            self.dictionary_on = True
            self.import_box.setVisible(True)
            self.dictionary = dictionary.get_dictionary()
            self.import_dictionary()
        else:
            self.dictionary_on = False
            self.dictionary_text = []
            self.import_box.setVisible(False)

if __name__ == '__main__':
    app = QApplication([])
    widget = OWQuerySearch()
    widget.show()
    app.exec()
    widget.saveSettings()





import numpy as np
from Orange.widgets.settings import Setting

from AnyQt.QtWidgets import QCheckBox

from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.widget import OWWidget, Msg

from Orange.widgets import gui

from orangecontrib.text import Corpus


class OWPosFilter(OWWidget):
    name = "POS filter"
    description = "Filter tokens on POS tags"
    icon = "icons/filter.svg"
    priority = 10

    want_main_area = False
    resizing_enabled = True

    pos_options = []
    pos_i = []
    pos = Setting([])

    drop_tag = Setting(True)
    drop_empty_doc = Setting(True)

    class Inputs:
        in_corpus = Input("Corpus", Corpus)

    class Outputs:
        out_corpus = Output("Corpus", Corpus)

    class Error(OWWidget.Error):
        no_tokens = Msg('No tokens left')

    def __init__(self):
        super().__init__()

        self.corpus = None

        gui.listBox(self.controlArea, self, 'pos_i', labels='pos_options', box = 'POS tags', selectionMode=2)
        gui.button(self.controlArea, self, 'Drop tag', toggleButton=True, value='drop_tag', buttonType=QCheckBox)
        gui.button(self.controlArea, self, 'Drop empty documents', toggleButton=True, value='drop_empty_doc', buttonType=QCheckBox)

        self.cs = gui.label(self.controlArea, self, 'Remembered selection: %(pos)s')
        self.cs.setWordWrap(True);

        gui.button(self.controlArea, self, 'Filter', self.filter_pos)


    def progress_with_info(self, n_done, n_all):
        self.progressBarSet(100 * (n_done / n_all if n_all else 1), None)  # prevent division by 0

    def filter_pos(self):
        self.pos = [self.pos_options[i] for i in self.pos_i]
        valid_docs = []
        if self.corpus.pos_tags is not None:
            out = self.corpus.copy()
            out._tokens = self.corpus._tokens.copy()
            out.pos_tags = self.corpus.pos_tags.copy()

            self.progressBarInit(None)
            n = len(self.corpus.pos_tags)
            for i, d in enumerate(self.corpus.pos_tags):

                selected = []
                for j, p in enumerate(d):
                    if p in self.pos:
                        selected.append(j)

                if len(selected) > 0:
                    valid_docs.append(i)
                out._tokens[i] = [self.corpus._tokens[i][j] for j in selected]
                if not self.drop_tag:
                    out.pos_tags[i] = [self.corpus.pos_tags[i][j] for j in selected]
                self.progress_with_info(i, n)

            if self.drop_tag:
                out.pos_tags = None

            if self.drop_empty_doc:
                out = out[valid_docs]

            self.progressBarFinished(None)


        if len(valid_docs) == 0:
            self.Error.no_tokens()
            self.Outputs.out_corpus.send(None)
        else:
            self.Error.no_tokens.clear()
            self.Outputs.out_corpus.send(out)


    def get_pos_options(self):
        pos = set()
        if self.corpus.pos_tags is not None:
            for d in self.corpus.pos_tags:
                pos = pos.union(np.unique(d))
        self.pos_options = sorted(list(pos))
        self.pos_i = [i for i,v in enumerate(self.pos_options) if v in self.pos]
        self.cs.setVisible(len(self.pos_options) == 0)

    @Inputs.in_corpus
    def set_data(self, in_corpus):
        self.corpus = in_corpus
        if self.corpus:
            self.get_pos_options()
            self.filter_pos()
            return
        self.Outputs.out_corpus.send(None)



if __name__ == '__main__':
    None